﻿using Microsoft.SqlServer.Server;
using System;
using System.Collections;
using System.Data.SqlClient;
using System.Data.SqlTypes;
using System.IO;
using System.Text;
using System.Globalization;

/*
Very important - use template from .NET Framework (not a standard or universal)
*/

namespace SunClr_Namespace  //namespace is not neccessary
{
    public class SunClr_Class
    {
        // FUNCTION TO SPIL TEXT INTO ROWS
        // example from https://www.skylinetechnologies.com/Blog/Skyline-Blog/March-2013/CLR-Functions-in-SQL-Server-A-Tutorial

        [SqlFunction(
            DataAccess = DataAccessKind.None,
            FillRowMethodName = "FillRowMethod"
            , IsDeterministic = true)
        ]
        //SQL Functions must be declared as Static. Table Valued functions must also
        //return a class that implements the IEnumerable interface. Most built in
        //.NET collections and arrays already implement this interface.
        public static IEnumerable Split(string stringToSplit, string delimiters)
        {
            //One line of C# code splits our string on one or more delimiters...
            //A string array is one of many objects that are returnable from
            //a SQL CLR function - as it implements the required IEnumerable interface.
            string[] elements = stringToSplit.Split(delimiters.ToCharArray());
            return elements;
        }
        //SQL needs to defer to user code to translate the an IEnumerable item into something
        //SQL Server can understand. In this case we convert our string to a SqlChar object...
        public static void FillRowMethod(Object theItem, out SqlChars results)
        {
            results = new SqlChars(theItem.ToString());
        }

        // Function listing files in a directory
         
        [SqlFunction(FillRowMethodName = "ListFiles_GetRow")]
        public static IEnumerable ListFiles(string dirPath)
        {
            DirectoryInfo dirInfo = new DirectoryInfo(dirPath);
            return dirInfo.GetFiles();
        }

        public static void ListFiles_GetRow(Object obj, out SqlString name, out SqlDateTime lastWriteDateTime, out SqlString type)
        {
            FileInfo fileInfo = (FileInfo)obj;
            name = new SqlString(fileInfo.Name);
            lastWriteDateTime = new SqlDateTime(fileInfo.LastWriteTime);
            type = new SqlString("FILE");
        }

        //  Function listing subdirectories in a directory
         
        [SqlFunction(FillRowMethodName = "ListDirs_GetRow")]
        public static IEnumerable ListDirs(string dirPath)
        {
            DirectoryInfo dirInfo = new DirectoryInfo(dirPath);
            return dirInfo.GetDirectories();
        }

        public static void ListDirs_GetRow(Object obj, out SqlString name, out SqlDateTime lastWriteDateTime, out SqlString type)
        {
            DirectoryInfo dirInfo = (DirectoryInfo)obj;
            name = new SqlString(dirInfo.Name);
            lastWriteDateTime = new SqlDateTime(dirInfo.LastWriteTime);
            type = new SqlString("DIR");
        }

        // Scalar function using internal connection to SQL 

        [SqlFunction(DataAccess = DataAccessKind.Read,
            SystemDataAccess = SystemDataAccessKind.Read)]
        public static int GetDBCount()
        {
            using (SqlConnection conn
                = new SqlConnection("context connection=true"))
            {
                conn.Open();
                SqlCommand cmd = new SqlCommand(
                    "SELECT COUNT(*) AS 'NumberOfDatabases' FROM sys.databases", conn);
                return (int)cmd.ExecuteScalar();
            }
        }

        //PROCEDURE

        [Microsoft.SqlServer.Server.SqlProcedure]
        public static void GetWindowsLocale()
        {
            CultureInfo ci = CultureInfo.InstalledUICulture;

            SqlContext.Pipe.Send(string.Format("Name: {0}", ci.Name));
            SqlContext.Pipe.Send(string.Format("Display Name: {0}", ci.DisplayName));
            SqlContext.Pipe.Send(string.Format("English Name: {0}", ci.EnglishName));
            SqlContext.Pipe.Send(string.Format("2-letter ISO Name: {0}", ci.TwoLetterISOLanguageName));
            SqlContext.Pipe.Send(string.Format("3-letter ISO Name: {0}", ci.ThreeLetterISOLanguageName));
            SqlContext.Pipe.Send(string.Format("3-letter Win32 API Name: {0}", ci.ThreeLetterWindowsLanguageName));
                                
            using (SqlConnection connection = new SqlConnection("context connection=true"))
            {
                connection.Open();
                SqlCommand command = new SqlCommand("SELECT SERVERPROPERTY('Collation') AS Collation", connection);
                SqlDataReader reader = command.ExecuteReader();
                SqlContext.Pipe.Send(reader);
            }
        }

        //TRIGGER DDL

        public static void DropTableTrigger()
        {
            SqlTriggerContext triggContext = SqlContext.TriggerContext;

            switch (triggContext.TriggerAction)
            {
                case TriggerAction.DropTable:
                    SqlContext.Pipe.Send("Table dropped! Here's the EventData:");
                    SqlContext.Pipe.Send(triggContext.EventData.Value);
                    break;

                default:
                    SqlContext.Pipe.Send("Something happened! Here's the EventData:");
                    SqlContext.Pipe.Send(triggContext.EventData.Value);
                    break;
            }
        }
    }

    // AGGREGATE - note this is a separate structure (not component of the class above

    [Serializable]
    [SqlUserDefinedAggregate(
        Format.Native,
        IsInvariantToDuplicates = false,
        IsInvariantToNulls = true,
        IsInvariantToOrder = true,
        IsNullIfEmpty = true,
        Name = "WeightedAvg")]
    public struct WeightedAvg
    {
        /// <summary>  
        /// The variable that holds the intermediate sum of all values multiplied by their weight  
        /// </summary>  
        private long sum;

        /// <summary>  
        /// The variable that holds the intermediate sum of all weights  
        /// </summary>  
        private int count;

        /// <summary>  
        /// Initialize the internal data structures  
        /// </summary>  
        public void Init()
        {
            sum = 0;
            count = 0;
        }

        /// <summary>  
        /// Accumulate the next value, not if the value is null  
        /// </summary>  
        /// <param name="Value">Next value to be aggregated</param>  
        /// <param name="Weight">The weight of the value passed to Value parameter</param>  
        public void Accumulate(SqlInt32 Value, SqlInt32 Weight)
        {
            if (!Value.IsNull && !Weight.IsNull)
            {
                sum += (long)Value * (long)Weight;
                count += (int)Weight;
            }
        }

        /// <summary>  
        /// Merge the partially computed aggregate with this aggregate  
        /// </summary>  
        /// <param name="Group">The other partial results to be merged</param>  
        public void Merge(WeightedAvg Group)
        {
            sum += Group.sum;
            count += Group.count;
        }

        /// <summary>  
        /// Called at the end of aggregation, to return the results of the aggregation.  
        /// </summary>  
        /// <returns>The weighted average of all inputed values</returns>  
        public SqlInt32 Terminate()
        {
            if (count > 0)
            {
                int value = (int)(sum / count);
                return new SqlInt32(value);
            }
            else
            {
                return SqlInt32.Null;
            }
        }
    }

    // TYPE - POINT  -  note this is a separate struct - not component of the class above


    [Serializable]
    [Microsoft.SqlServer.Server.SqlUserDefinedType(Format.Native,
         IsByteOrdered = true, ValidationMethodName = "ValidatePoint")]
    public struct Point : INullable
    {
        private bool is_Null;
        private Int32 _x;
        private Int32 _y;

        public bool IsNull
        {
            get
            {
                return (is_Null);
            }
        }

        public static Point Null
        {
            get
            {
                Point pt = new Point();
                pt.is_Null = true;
                return pt;
            }
        }

        // Use StringBuilder to provide string representation of UDT.  
        public override string ToString()
        {
            // Since InvokeIfReceiverIsNull defaults to 'true'  
            // this test is unneccesary if Point is only being called  
            // from SQL.  
            if (this.IsNull)
                return "NULL";
            else
            {
                StringBuilder builder = new StringBuilder();
                builder.Append(_x);
                builder.Append(",");
                builder.Append(_y);
                return builder.ToString();
            }
        }

        [SqlMethod(OnNullCall = false)]
        public static Point Parse(SqlString s)
        {
            // With OnNullCall=false, this check is unnecessary if   
            // Point only called from SQL.  
            if (s.IsNull)
                return Null;

            // Parse input string to separate out points.  
            Point pt = new Point();
            string[] xy = s.Value.Split(",".ToCharArray());
            pt.X = Int32.Parse(xy[0]);
            pt.Y = Int32.Parse(xy[1]);

            // Call ValidatePoint to enforce validation  
            // for string conversions.  
            if (!pt.ValidatePoint())
                throw new ArgumentException("Invalid XY coordinate values.");
            return pt;
        }

        // X and Y coordinates exposed as properties.  
        public Int32 X
        {
            get
            {
                return this._x;
            }
            // Call ValidatePoint to ensure valid range of Point values.  
            set
            {
                Int32 temp = _x;
                _x = value;
                if (!ValidatePoint())
                {
                    _x = temp;
                    throw new ArgumentException("Invalid X coordinate value.");
                }
            }
        }

        public Int32 Y
        {
            get
            {
                return this._y;
            }
            set
            {
                Int32 temp = _y;
                _y = value;
                if (!ValidatePoint())
                {
                    _y = temp;
                    throw new ArgumentException("Invalid Y coordinate value.");
                }
            }
        }

        // Validation method to enforce valid X and Y values.  
        private bool ValidatePoint()
        {
            // Allow only zero or positive integers for X and Y coordinates.  
            if ((_x >= 0) && (_y >= 0))
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        // Distance from 0 to Point method.  
        [SqlMethod(OnNullCall = false)]
        public Double Distance()
        {
            return DistanceFromXY(0, 0);
        }

        // Distance from Point to the specified point method.  
        [SqlMethod(OnNullCall = false)]
        public Double DistanceFrom(Point pFrom)
        {
            return DistanceFromXY(pFrom.X, pFrom.Y);
        }

        // Distance from Point to the specified x and y values method.  
        [SqlMethod(OnNullCall = false)]
        public Double DistanceFromXY(Int32 iX, Int32 iY)
        {
            return Math.Sqrt(Math.Pow(iX - _x, 2.0) + Math.Pow(iY - _y, 2.0));
        }
    }
}
